@@ -2,7 +2,8 @@ package gojq
 
 import (
 	"context"
-	"fmt"
+	"math"
+	"reflect"
 	"sort"
 )
 
@@ -42,9 +43,9 @@ loop:
 		case oppop:
 			env.pop()
 		case opdup:
-			x := env.pop()
-			env.push(x)
-			env.push(x)
+			v := env.pop()
+			env.push(v)
+			env.push(v)
 		case opconst:
 			env.pop()
 			env.push(code.v)
@@ -78,9 +79,8 @@ loop:
 				}
 				pc, backtrack = code.v.(int), false
 				goto loop
-			} else {
-				env.pushfork(pc)
 			}
+			env.pushfork(pc)
 		case opforktrybegin:
 			if backtrack {
 				if err == nil {
@@ -107,18 +107,16 @@ loop:
 				}
 				pc, backtrack, err = code.v.(int), false, nil
 				goto loop
-			} else {
-				env.pushfork(pc)
 			}
+			env.pushfork(pc)
 		case opforktryend:
 			if backtrack {
 				if err != nil {
 					err = &tryEndError{err}
 				}
 				break loop
-			} else {
-				env.pushfork(pc)
 			}
+			env.pushfork(pc)
 		case opforkalt:
 			if backtrack {
 				if err == nil {
@@ -126,9 +124,8 @@ loop:
 				}
 				pc, backtrack, err = code.v.(int), false, nil
 				goto loop
-			} else {
-				env.pushfork(pc)
 			}
+			env.pushfork(pc)
 		case opforklabel:
 			if backtrack {
 				label := env.pop()
@@ -136,13 +133,12 @@ loop:
 					err = nil
 				}
 				break loop
-			} else {
-				env.push(env.label)
-				env.pushfork(pc)
-				env.pop()
-				env.values[env.index(code.v.([2]int))] = env.label
-				env.label++
 			}
+			env.push(env.label)
+			env.pushfork(pc)
+			env.pop()
+			env.values[env.index(code.v.([2]int))] = env.label
+			env.label++
 		case opbacktrack:
 			break loop
 		case opjump:
@@ -153,6 +149,30 @@ loop:
 				pc = code.v.(int)
 				goto loop
 			}
+		case opindex, opindexarray:
+			if backtrack {
+				break loop
+			}
+			p, v := code.v, env.pop()
+			if code.op == opindexarray && v != nil {
+				if _, ok := v.([]interface{}); !ok {
+					err = &expectedArrayError{v}
+					break loop
+				}
+			}
+			w := funcIndex2(nil, v, p)
+			if e, ok := w.(error); ok {
+				err = e
+				break loop
+			}
+			env.push(w)
+			if !env.paths.empty() && env.expdepth == 0 {
+				if !env.pathIntact(v) {
+					err = &invalidPathError{v}
+					break loop
+				}
+				env.paths.push(pathValue{path: p, value: w})
+			}
 		case opcall:
 			if backtrack {
 				break loop
@@ -175,14 +195,31 @@ loop:
 					break loop
 				}
 				env.push(w)
-				if !env.paths.empty() {
-					var ps []interface{}
-					ps, err = env.pathEntries(v[2].(string), x, args)
-					if err != nil {
-						break loop
-					}
-					for _, p := range ps {
-						env.paths.push(pathValue{path: p, value: w})
+				if !env.paths.empty() && env.expdepth == 0 {
+					switch v[2].(string) {
+					case "_index":
+						if x = args[0]; !env.pathIntact(x) {
+							err = &invalidPathError{x}
+							break loop
+						}
+						env.paths.push(pathValue{path: args[1], value: w})
+					case "_slice":
+						if x = args[0]; !env.pathIntact(x) {
+							err = &invalidPathError{x}
+							break loop
+						}
+						env.paths.push(pathValue{
+							path:  map[string]interface{}{"start": args[2], "end": args[1]},
+							value: w,
+						})
+					case "getpath":
+						if !env.pathIntact(x) {
+							err = &invalidPathError{x}
+							break loop
+						}
+						for _, p := range args[0].([]interface{}) {
+							env.paths.push(pathValue{path: p, value: w})
+						}
 					}
 				}
 			default:
@@ -199,7 +236,7 @@ loop:
 			goto loop
 		case opscope:
 			xs := code.v.([3]int)
-			var saveindex, outerindex, limit int
+			var saveindex, outerindex int
 			if index == env.scopes.index {
 				if callpc >= 0 {
 					saveindex = index
@@ -207,7 +244,7 @@ loop:
 					callpc, saveindex = env.popscope()
 				}
 			} else {
-				env.scopes.save(&saveindex, &limit)
+				saveindex, _ = env.scopes.save()
 				env.scopes.index = index
 			}
 			if outerindex = index; outerindex >= 0 {
@@ -230,7 +267,7 @@ loop:
 			if env.scopes.empty() {
 				return env.pop(), true
 			}
-		case opeach:
+		case opiter:
 			if err != nil {
 				break loop
 			}
@@ -240,8 +277,7 @@ loop:
 			case []pathValue:
 				xs = v
 			case []interface{}:
-				if !env.paths.empty() && env.expdepth == 0 &&
-					!deepEqual(v, env.paths.top().(pathValue).value) {
+				if !env.paths.empty() && env.expdepth == 0 && !env.pathIntact(v) {
 					err = &invalidPathIterError{v}
 					break loop
 				}
@@ -253,8 +289,7 @@ loop:
 					xs[i] = pathValue{path: i, value: v}
 				}
 			case map[string]interface{}:
-				if !env.paths.empty() && env.expdepth == 0 &&
-					!deepEqual(v, env.paths.top().(pathValue).value) {
+				if !env.paths.empty() && env.expdepth == 0 && !env.pathIntact(v) {
 					err = &invalidPathIterError{v}
 					break loop
 				}
@@ -289,6 +324,7 @@ loop:
 				break loop
 			default:
 				err = &iteratorError{v}
+				env.push(emptyIter{})
 				break loop
 			}
 			if len(xs) > 1 {
@@ -312,18 +348,13 @@ loop:
 			if backtrack {
 				break loop
 			}
-			if env.expdepth > 0 {
-				panic(fmt.Sprintf("unexpected expdepth: %d", env.expdepth))
-			}
 			env.pop()
-			x := env.pop()
-			if deepEqual(x, env.paths.top().(pathValue).value) {
-				env.push(env.poppaths())
-				env.expdepth = env.paths.pop().(int)
-			} else {
-				err = &invalidPathError{x}
+			if v := env.pop(); !env.pathIntact(v) {
+				err = &invalidPathError{v}
 				break loop
 			}
+			env.push(env.poppaths())
+			env.expdepth = env.paths.pop().(int)
 		default:
 			panic(code.op)
 		}
@@ -357,9 +388,9 @@ func (env *env) popscope() (int, int) {
 
 func (env *env) pushfork(pc int) {
 	f := fork{pc: pc, expdepth: env.expdepth}
-	env.stack.save(&f.stackindex, &f.stacklimit)
-	env.scopes.save(&f.scopeindex, &f.scopelimit)
-	env.paths.save(&f.pathindex, &f.pathlimit)
+	f.stackindex, f.stacklimit = env.stack.save()
+	f.scopeindex, f.scopelimit = env.scopes.save()
+	f.pathindex, f.pathlimit = env.paths.save()
 	env.forks = append(env.forks, f)
 	env.debugForks(pc, ">>>")
 }
@@ -389,32 +420,21 @@ type pathValue struct {
 	path, value interface{}
 }
 
-func (env *env) pathEntries(name string, x interface{}, args []interface{}) ([]interface{}, error) {
-	switch name {
-	case "_index":
-		if env.expdepth > 0 {
-			return nil, nil
-		} else if !deepEqual(args[0], env.paths.top().(pathValue).value) {
-			return nil, &invalidPathError{x}
+func (env *env) pathIntact(v interface{}) bool {
+	w := env.paths.top().(pathValue).value
+	switch v := v.(type) {
+	case []interface{}, map[string]interface{}:
+		switch w.(type) {
+		case []interface{}, map[string]interface{}:
+			v, w := reflect.ValueOf(v), reflect.ValueOf(w)
+			return v.Pointer() == w.Pointer() && v.Len() == w.Len()
 		}
-		return []interface{}{args[1]}, nil
-	case "_slice":
-		if env.expdepth > 0 {
-			return nil, nil
-		} else if !deepEqual(args[0], env.paths.top().(pathValue).value) {
-			return nil, &invalidPathError{x}
+	case float64:
+		if w, ok := w.(float64); ok {
+			return v == w || math.IsNaN(v) && math.IsNaN(w)
 		}
-		return []interface{}{map[string]interface{}{"start": args[2], "end": args[1]}}, nil
-	case "getpath":
-		if env.expdepth > 0 {
-			return nil, nil
-		} else if !deepEqual(x, env.paths.top().(pathValue).value) {
-			return nil, &invalidPathError{x}
-		}
-		return args[0].([]interface{}), nil
-	default:
-		return nil, nil
 	}
+	return v == w
 }
 
 func (env *env) poppaths() []interface{} {
