@@ -4,7 +4,6 @@ import (
 	"encoding/json"
 	"errors"
 	"fmt"
-	"math/big"
 	"reflect"
 	"testing"
 
@@ -37,9 +36,12 @@ func TestWithModuleLoaderError(t *testing.T) {
 	if !ok {
 		t.Fatal("should emit an error but got no output")
 	}
-	err, expected := v.(error), `cannot load module: "m"`
-	if got := err.Error(); got != expected {
-		t.Errorf("expected: %v, got: %v", expected, got)
+	if err, ok := v.(error); ok {
+		if expected := `cannot load module: "m"`; err.Error() != expected {
+			t.Errorf("expected: %v, got: %v", expected, err)
+		}
+	} else {
+		t.Errorf("should emit an error but got: %v", v)
 	}
 	v, ok = iter.Next()
 	if ok {
@@ -96,7 +98,7 @@ func (*moduleLoaderJSON) LoadJSON(name string) (interface{}, error) {
 func TestWithModuleLoader_JSON(t *testing.T) {
 	query, err := gojq.Parse(`
 		import "module1" as $m;
-		[$m, $m[1]*$m[2]*1000000000000]
+		[$m, $m[1]*$m[2]]
 	`)
 	if err != nil {
 		t.Fatal(err)
@@ -114,10 +116,7 @@ func TestWithModuleLoader_JSON(t *testing.T) {
 		if !ok {
 			break
 		}
-		if expected := []interface{}{
-			[]interface{}{1.0, 42, 123},
-			big.NewInt(5166000000000000),
-		}; !reflect.DeepEqual(got, expected) {
+		if expected := []interface{}{[]interface{}{1.0, 42, 123}, 5166}; !reflect.DeepEqual(got, expected) {
 			t.Errorf("expected: %v, got: %v", expected, got)
 		}
 	}
@@ -182,7 +181,7 @@ func TestWithEnvironLoader(t *testing.T) {
 		}
 		expected := map[string]interface{}{"foo": "42", "bar": "128", "qux": ""}
 		if !reflect.DeepEqual(got, expected) {
-			t.Errorf("expected: %#v, got: %#v", expected, got)
+			t.Errorf("expected: %v, got: %v", expected, got)
 		}
 	}
 }
@@ -244,9 +243,12 @@ func TestWithVariablesError1(t *testing.T) {
 	if !ok {
 		t.Fatal("should emit an error but got no output")
 	}
-	err, expected := v.(error), "variable defined but not bound: $x"
-	if got := err.Error(); got != expected {
-		t.Errorf("expected: %v, got: %v", expected, got)
+	if err, ok := v.(error); ok {
+		if expected := "variable defined but not bound: $x"; err.Error() != expected {
+			t.Errorf("expected: %v, got: %v", expected, err)
+		}
+	} else {
+		t.Errorf("should emit an error but got: %v", v)
 	}
 	v, ok = iter.Next()
 	if ok {
@@ -271,9 +273,12 @@ func TestWithVariablesError2(t *testing.T) {
 	if !ok {
 		t.Fatal("should emit an error but got no output")
 	}
-	err, expected := v.(error), "too many variable values provided"
-	if got := err.Error(); got != expected {
-		t.Errorf("expected: %v, got: %v", expected, got)
+	if err, ok := v.(error); ok {
+		if expected := "too many variable values provided"; err.Error() != expected {
+			t.Errorf("expected: %v, got: %v", expected, err)
+		}
+	} else {
+		t.Errorf("should emit an error but got: %v", v)
 	}
 	v, ok = iter.Next()
 	if ok {
@@ -543,7 +548,7 @@ func TestWithFunctionValueError(t *testing.T) {
 			break
 		}
 		if !reflect.DeepEqual(v, expected) {
-			t.Errorf("expected: %#v, got: %#v", expected, v)
+			t.Errorf("expected: %v, got: %v", expected, v)
 		}
 	}
 }
@@ -710,9 +715,12 @@ func TestWithIterFunctionPathError(t *testing.T) {
 	if !ok {
 		t.Fatal("should emit an error but got no output")
 	}
-	err, expected := v.(error), "invalid path on iterating against: gojq.Iter"
-	if got := err.Error(); got != expected {
-		t.Errorf("expected: %v, got: %v", expected, got)
+	if err, ok := v.(error); ok {
+		if expected := "invalid path on iterating against: gojq.Iter"; err.Error() != expected {
+			t.Errorf("expected: %v, got: %v", expected, err)
+		}
+	} else {
+		t.Errorf("should emit an error but got: %v", v)
 	}
 }
 
